// Copyright (C) 2011 Nico Piatkowski <nico.piatkowski@tu-dortmund.de>, All rights reserved 
#ifndef ObservationSet_H
#define ObservationSet_H

#include <vector>
#include <set>
#include <string>
#include <iostream>
#include <fstream>
#include <sstream>
#include <algorithm>

#include <vector_types.h>

#include "AbstractLayer/AbstractGraph.h"
#include "AbstractLayer/AttributeConsumer.h"

#include "StructuredObservation.h"

#include <ext/hash_map>

template <class T>
class ObservationSet{
	private:
		std::vector<StructuredObservation<T>*> observations;
		std::vector<unsigned int> structureSizes;

		__gnu_cxx::hash_map<ParameterDescriptor, int>* parameterCounts;
		__gnu_cxx::hash_map<unsigned int, std::vector<int2>* >* attObservationMap;

		unsigned int _obsCount;
		unsigned int _length;

		Dictionary* attDict;
		Dictionary* classDict;

		unsigned int maxAtt;

		bool updateDict;

		bool updateObservations;
		bool updateNodeMap;
		bool updateStructureMap;
		bool updateLabels;
		bool updateValues;

		unsigned int nodeMapSize;
		unsigned int observationSize;

		unsigned int totalNodes;
		unsigned int totalNodesInFactors;
		unsigned int maxNodes;

		unsigned int* structureMap;
		unsigned int* nodeMap;
		unsigned int* obs;
		unsigned int* labels;
		T* values;

		AttributeConsumer<T>* aC;

	public:

		void setPCounts(__gnu_cxx::hash_map<ParameterDescriptor, int>* pC){
			parameterCounts = pC;
		}

		void setAOMap(__gnu_cxx::hash_map<unsigned int, std::vector<int2>* >* aOM){
			attObservationMap = aOM;
		}

		__gnu_cxx::hash_map<ParameterDescriptor, int>* getPCounts(){
			return parameterCounts;
		}

		__gnu_cxx::hash_map<unsigned int, std::vector<int2>* >* getAOMap(){
			return attObservationMap;
		}

		void setOCount(unsigned int oC){
			_obsCount	= oC;
		}

		unsigned int getOCount(){
			return _obsCount;
		}

		void setLen(unsigned int l){
			_length		= l;
		}

		unsigned int getLen(){
			return _length;
		}

		ObservationSet(AttributeConsumer<T>* ac){
			//std::cout << "ObservationSet<float>()" << std::endl;

			aC = ac;

			nodeMapSize     = 0;
			observationSize = 0;
			maxAtt			= 0;
			maxNodes		= 0;

			totalNodes = 0;
			totalNodesInFactors = 0;

			attDict    = new Dictionary();
			classDict  = new Dictionary();

			//std::string* O = new std::string("O");
			//classDict->addToken(O);

			updateDict = true;

			updateObservations	= false;
			updateNodeMap		= false;
			updateStructureMap	= false;
			updateLabels        = false;
			updateValues        = false;

			structureMap = NULL;
			nodeMap      = NULL;
			obs          = NULL;
			labels       = NULL;
			values		 = NULL;

			_obsCount	= 0;
			_length		= 0;

			parameterCounts   = NULL;
			attObservationMap = NULL;
		};

		ObservationSet(Dictionary* a, Dictionary* c, AttributeConsumer<T>* ac, bool u = false){
			//std::cout << "ObservationSet<float>(" << a << ", " << c << ")" << std::endl;

			aC = ac;

			nodeMapSize     = 0;
			observationSize = 0;
			maxAtt			= 0;
			maxNodes		= 0;

			totalNodes = 0;
			totalNodesInFactors	= 0;

			attDict    = a;
			classDict  = c;

			updateDict = u;

			updateObservations	= false;
			updateNodeMap		= false;
			updateStructureMap	= false;
			updateLabels        = false;
			updateValues        = false;

			structureMap = NULL;
			nodeMap      = NULL;
			obs          = NULL;
			labels       = NULL;
			values		 = NULL;

			_obsCount	= 0;
			_length		= 0;

			parameterCounts   = NULL;
			attObservationMap = NULL;
		};

		~ObservationSet(){
			if(updateDict){
				delete attDict;
				delete classDict;
			}

			if(structureMap) free(structureMap);
			if(nodeMap) free(nodeMap);
			if(obs) free(obs);
			if(labels) free(labels);
			if(values) free(values);

			typename std::vector<StructuredObservation<T>* >::iterator ii = observations.begin();

			for(; ii != observations.end(); ii++){
				delete *ii;
			}
		};

		void clear(){

			updateObservations	= true;
			updateNodeMap		= true;
			updateStructureMap	= true;
			updateLabels        = true;
			updateValues		= true;

			typename std::vector<StructuredObservation<T>* >::iterator ii = observations.begin();
			for(; ii != observations.end(); ii++){
				delete *ii;
			}
			observations.clear();

			nodeMapSize     = 0;
			observationSize = 0;
			maxAtt			= 0;

			totalNodes = 0;
			totalNodesInFactors	= 0;

			structureSizes.clear();
		};

		unsigned int getTotalNodesInFactors(){
			return totalNodesInFactors;
		}

		void blankLabels(){
			for(unsigned int i=0; i<nodeMapSize; i++){
				labels[i] = 0;
			}
		}

		unsigned int getMaxSize(){
			return *std::max_element(structureSizes.begin(), structureSizes.end());
		}

		void add(StructuredObservation<T>* s){

			//std::cout << "Beispiel " << (observations.size()+1) << std::endl;

			observations.push_back(s);
			structureSizes.push_back(s->getNodeCount());

			nodeMapSize			+= s->getNodeCount() + s->getNilCount();
			observationSize		+= s->getSize();
			totalNodesInFactors	+= s->getNodesInFactors();

			updateObservations	= true;
			updateNodeMap		= true;
			updateStructureMap	= true;
			updateLabels        = true;
		};

		StructuredObservation<T>* get(unsigned int i){
			return observations[i];
		}

		unsigned int getSize(){
			return observations.size();
		};

		unsigned int getLabelCount(){
			return classDict->getSize();
		}

		unsigned int getAttCount(){
			return attDict->getSize();
		}


		unsigned int getLastAttributeIndex(){
			return maxAtt;
		}

		// unsigned int readFromFile(std::string fName)
		//
		// Liest Daten im Format:
		// LABEL ATTRIBUT_1 ATTRIBUT_2 ...
		// ..
		// LABEL ATTRIBUT_1 ATTRIBUT_2 ...
		// [Leerzeile]
		// LABEL ATTRIBUT_1 ATTRIBUT_2 ...
		// ..
		// LABEL ATTRIBUT_1 ATTRIBUT_2 ...
		// [Leerzeile]
		// ..
		// Beobachtungen sind durch eine Leerzeile voneinander getrennt.
		// Innerhalb einer Beobachtung entspricht die i-te Zeile dem
		// Label sowie den Attributausprägungen an Knoten i.
		//
		// Erzeugt die Trainingsmenge T := (X,Y)
		//
		unsigned int readFromFile(std::string* fName, int toRead = -1){

			ifstream dataFile(fName->c_str());
			const unsigned int b = readFromStream(&dataFile, toRead);
			dataFile.close();

			return b;
		};

		unsigned int getMaxNodeCount(){
			return maxNodes;
		}

		unsigned int readFromStream(std::istream* is, int toRead = -1){

			unsigned int read = 0;
			unsigned int bytes = 0;

			std::string s;
			StructuredObservation<T>* o = new StructuredObservation<T>(attDict, classDict, updateDict);

			while(read < toRead && !is->eof()){
				std::getline(*is, s);
				bytes += s.size();

				if(s.length()<2){
					if(o->getNodeCount()>0){ // && o->getNodeCount()<(AbstractGraph::pbase-1)){
						if(o->getNodeCount()>maxNodes) maxNodes = o->getNodeCount();
						assert(maxNodes < AbstractGraph::pbase);
						add(o);
						read++;
					}
					else delete o;
					o = new StructuredObservation<T>(attDict, classDict, updateDict);
				}
				else{
					if(o->addNode(&s, maxAtt, aC, totalNodes, getSize())) totalNodes++;
				}

			}

			delete o;
			return bytes;
		};

		unsigned int* getObservations(){
			if(updateObservations){
				unsigned int* O = NULL;
				T* V = NULL;
				unsigned int L = 0;

				typename std::vector<StructuredObservation<T>* >::iterator ii = observations.begin();

				for(; ii != observations.end(); ii++){
	
					L += (*ii)->getSize();

					if(!O){
						O = (unsigned int*)malloc( L*sizeof(unsigned int));
						V = (T*)malloc( L*sizeof(T));
					}
					else{
						O = (unsigned int*)realloc(O, L*sizeof(unsigned int));
						V = (T*)realloc(V, L*sizeof(T));
					}

					memcpy(&O[L-(*ii)->getSize()], (*ii)->getObservations(), (*ii)->getSize()*sizeof(unsigned int));
					memcpy(&V[L-(*ii)->getSize()], (*ii)->getValues(), (*ii)->getSize()*sizeof(T));
				}

				if(obs) free(obs);
				if(values) free(values);
				obs = O;
				values = V;
				updateObservations = false;
				return O;
			}else{
				return obs;
			}
		};

		T* getValues(){
			return values;
		};

		unsigned int* getLabels(){
			if(updateLabels){
				unsigned int* O = NULL;
				unsigned int L = 0;

				typename std::vector<StructuredObservation<T>* >::iterator ii = observations.begin();

				for(; ii != observations.end(); ii++){
	
					L += (*ii)->getNodeCount();

					if(!O){
						O = (unsigned int*)malloc( L*sizeof(unsigned int));
					}
					else{
						O = (unsigned int*)realloc(O, L*sizeof(unsigned int));
					}

					memcpy(&O[L-(*ii)->getNodeCount()], (*ii)->getLabels(), (*ii)->getNodeCount()*sizeof(unsigned int));
				}

				if(labels) free(labels);
				labels = O;
				updateLabels = false;
				return O;
			}else{
				return labels;
			}
		};

		unsigned int* getStructureMap(){
			if(updateStructureMap){
				unsigned int* M = (unsigned int*)malloc((getSize()+1)*sizeof(unsigned int));

				unsigned int i = 0, j = 0;

				typename std::vector<StructuredObservation<T>* >::iterator ii = observations.begin();

				for(; ii != observations.end(); ii++){
					M[i] = j;
					j += (*ii)->getNodeCount();
					i++;
				}
				M[getSize()] = j;

				if(structureMap) free(structureMap);
				structureMap = M;
				updateStructureMap = false;
				return M;
			}else{
				return structureMap;
			}
		};

		unsigned int* getNodeMap(){
			if(updateNodeMap){
				unsigned int* M = (unsigned int*)malloc((nodeMapSize+1)*sizeof(unsigned int));

				unsigned int h = 0, i = 0, j = 0;

				typename std::vector<StructuredObservation<T>* >::iterator ii = observations.begin();

				unsigned int* m = NULL;

				for(; ii != observations.end(); ii++){
					m = (*ii)->getNodeMap();
					for(j = 0; j<(*ii)->getNodeCount(); j++){
						M[i] = m[j] + h;
						i++;
					}
					h += m[j];
				}
				M[nodeMapSize] = h;

				if(nodeMap) free(nodeMap);
				nodeMap = M;
				updateNodeMap = false;
				return M;
			}else{
				return nodeMap;
			}
		};

		unsigned int* getStructureSizes(){
			return &structureSizes[0];
		}

		unsigned int getNodeMapSize(){
			return nodeMapSize+1;
		};

		unsigned int getStructureMapSize(){
			return getSize()+1;
		};

		unsigned int getObservationSize(){
			return observationSize;
		};

		void print(){
			std::string* s;

			typename std::vector<StructuredObservation<T>* >::iterator ii = observations.begin();

			for(; ii != observations.end(); ii++){
				s = (*ii)->toString();
				std::cout << (*s) << std::endl;
				delete s;
			}
		};

		Dictionary* getAttDict(){ return attDict; };
		Dictionary* getClassDict(){ return classDict; };
};
#endif
