// Copyright (C) 2011 Nico Piatkowski <nico.piatkowski@tu-dortmund.de>, All rights reserved 
#ifndef ForwardBackward_CUH
#define ForwardBackward_CUH

#include "ForwardBackwardConfig.h"
#include "LinearModelConfig.h"

/*********************************************************************/

template <class T>
__device__ T getAlpha(unsigned int structure, int t, unsigned int i, ForwardBackwardConfig<T>* cConfig, LinearModelConfig<T>* conf){
	return cConfig->alpha[((conf->structMap[structure] + t) * conf->labelCount) + i];
}

template <class T>
__device__ void setAlpha(unsigned int structure, int t, T value, ForwardBackwardConfig<T>* cConfig, LinearModelConfig<T>* conf){
	cConfig->alpha[((conf->structMap[structure] + t) * conf->labelCount) + threadIdx.x] = value;
}


template <class T>
__device__ T forwardStep(volatile T* sdata, unsigned int structure, unsigned int t, ForwardBackwardConfig<T>* cConfig, LinearModelConfig<T>* conf){

	const unsigned int propIdx = 2*conf->labelCount+(conf->labelCount*conf->labelCount);
	T q, p;

	bool active = (threadIdx.x == getLabel(structure, t, cConfig, conf));

	// ertse Iteration abgerollt
	if(active && 0 == getLabel(structure, t-1, cConfig, conf)){
		sdata[propIdx] += getLogEdgePotential<T>(sdata, 0, threadIdx.x, conf) + getLogNodePotential(sdata, threadIdx.x, conf);
		active = false;
	}

	q = sdata[0] + getLogEdgePotential<T>(sdata, 0, threadIdx.x, conf) + getLogNodePotential(sdata, threadIdx.x, conf);

	for(unsigned int i = 1; i < conf->labelCount; i++){
		p = sdata[i] + getLogEdgePotential<T>(sdata, i, threadIdx.x, conf) + getLogNodePotential(sdata, threadIdx.x, conf);
		q = logAdd(q, p);
		
		if(active && i == getLabel(structure, t-1, cConfig, conf)){
			sdata[propIdx] += getLogEdgePotential<T>(sdata, i, threadIdx.x, conf) + getLogNodePotential(sdata, threadIdx.x, conf);
			active = false;
		}
	}
	return q;
}

template <class T>
__device__ T backwardStep(volatile T* sdata, LinearModelConfig<T>* conf){
	T p, q = getLogEdgePotential<T>(sdata, threadIdx.x, 0, conf) + getLogNodePotential(sdata, 0, conf) + sdata[0];

	for(unsigned int j = 1; j < conf->labelCount; j++){
		p = getLogEdgePotential<T>(sdata, threadIdx.x, j, conf) + getLogNodePotential(sdata, j, conf) + sdata[j];
		q = logAdd(q,p);
	}
	return q;
}

template <class T>
__device__ unsigned int getLabel(unsigned int structure, int t, ForwardBackwardConfig<T>* cConfig, LinearModelConfig<T>* conf){
	return cConfig->labels[conf->structMap[structure] + t];
}

template <class T>
__global__ void fbLCRFKernel(unsigned int start, ForwardBackwardConfig<T>* cConfig, LinearModelConfig<T>* conf){

	// y_i := threadIdx.x
	const unsigned int structure = start + blockIdx.x;
	const unsigned int L = conf->structureSizes[structure];
	const unsigned int label = threadIdx.x;
	const unsigned int propIdx = 2*conf->labelCount+(conf->labelCount*conf->labelCount);

	volatile T *shared = SharedMemory<T>();
	T q, p, w, h, logZ = (T)0.0;
	if(threadIdx.x==0) shared[propIdx] = (T)0.0;

	clearPotentials(shared, conf);

	///////////////////////////////////////////////////////////////////////////////
	// Forward
	///////////////////////////////////////////////////////////////////////////////

	computeLogPotential<T>(shared, 0, structure, conf);
	p = getLogNodePotential(shared, label, conf);

	if(label == getLabel(structure, 0, cConfig, conf)) shared[propIdx] += (p);

	storeSync(shared, p);
	setAlpha(structure, 0, p, cConfig, conf);
	clearPotentials(shared, conf);

	for(unsigned int t=1; t<L; t++){

		computeLogPotential<T>(shared, t, structure, conf);
		q = forwardStep<T>(shared, structure, t, cConfig, conf);

		storeSync(shared, q);
		setAlpha(structure, t, q, cConfig, conf);
		clearPotentials(shared, conf);
	}

	logZ = shared[0];
	for(unsigned int i=1; i<blockDim.x; i++){
		logZ = logAdd(logZ, shared[i]);
	}

	if(label == getLabel(structure, L-1, cConfig, conf)){
		cConfig->logProps[structure] = shared[label] - logZ;
	}

	///////////////////////////////////////////////////////////////////////////////
	// Backward
	///////////////////////////////////////////////////////////////////////////////

	q = (T)0.0;
	w = (T)0.0;
	storeSync(shared, q);
	clearPotentials(shared, conf);

	for(int t=L-1; t>0; t--){

		p = getAlpha<T>(structure, t, threadIdx.x, cConfig, conf);

		setNodeMarginal<T>(structure, t, (T)exp(p+q-logZ), conf);
		computeLogPotential<T>(shared, t, structure, conf);
		h = getLogNodePotential(shared, label, conf);

		for(unsigned int i=0; i<blockDim.x; i++){
			w = getLogEdgePotential<T>(shared, i, threadIdx.x, conf);
			p = getAlpha<T>(structure, t-1, i, cConfig, conf);
			setEdgeMarginal<T>(structure, t, i, (T)exp(p+w+h+q-logZ), conf);
			
		}

		q = backwardStep<T>(shared, conf);
		storeSync<T>(shared, q);
		clearPotentials<T>(shared, conf);
	}

	p = getAlpha<T>(structure, 0, threadIdx.x, cConfig, conf);

	setNodeMarginal<T>(structure, 0, (T)exp(q+p-logZ), conf);
}

template <class T> void fbLCRFGPU(ForwardBackwardConfig<T>* conf, LinearModelConfig<T>* pConfig){

	const unsigned int threads   = pConfig->labelCount;
	const unsigned int smem      = (2*pConfig->labelCount+(pConfig->labelCount*pConfig->labelCount)+1)*sizeof(T);

	unsigned int totalTasks = pConfig->structureCount;
	unsigned int tasks      = min(conf->maxBlocks, totalTasks);
	unsigned int start      = 0;

	dim3 fbBlock(threads, 1, 1);

	LinearModelConfig<T>* gpuPConfig;
	cudaMalloc((void**)&gpuPConfig, sizeof(LinearModelConfig<T>));
	cudaMemcpy(gpuPConfig, pConfig, sizeof(LinearModelConfig<T>), cudaMemcpyHostToDevice);

	ForwardBackwardConfig<T>* gpuConfig;
	cudaMalloc((void**)&gpuConfig, sizeof(ForwardBackwardConfig<T>));
	cudaMemcpy(gpuConfig, conf, sizeof(ForwardBackwardConfig<T>), cudaMemcpyHostToDevice);

	//cudaPrintfInit(64*1048576);

	while(totalTasks){
		dim3 fbGrid(tasks, 1, 1);

		fbLCRFKernel<T><<<fbGrid, fbBlock, smem>>>(start, gpuConfig, gpuPConfig);

		start += tasks;
		totalTasks -= tasks;
		tasks = min(conf->maxBlocks, totalTasks);
	}
	//cudaPrintfDisplay(stdout, true);
	//cudaPrintfEnd();

	cudaFree(gpuConfig);
	cudaFree(gpuPConfig);
}

template void fbLCRFGPU<float>(ForwardBackwardConfig<float>*, LinearModelConfig<float>*);
template void fbLCRFGPU<double>(ForwardBackwardConfig<double>*, LinearModelConfig<double>*);

///////////////////////////////////////////////////////////////////////////////
// Viterbi-Kernel
///////////////////////////////////////////////////////////////////////////////

template <class T>
__device__ void setViterbiPath(unsigned int structure, int t, unsigned int idx, ForwardBackwardConfig<T>* conf, LinearModelConfig<T>* pConfig){
	conf->viterbiPath[(pConfig->structMap[structure] + t)*blockDim.x+threadIdx.x] = idx;
}

template <class T>
__global__ void viterbiLCRFKernel(unsigned int start, ForwardBackwardConfig<T>* cConfig, LinearModelConfig<T>* conf){

	// y_i := threadIdx.x
	const unsigned int structure = start + blockIdx.x;
	const unsigned int L = conf->structureSizes[structure];
	const unsigned int label = threadIdx.x;
	unsigned int maxIdx = threadIdx.x;

	volatile T *shared = SharedMemory<T>();
	T q, p, maxValue;

	clearPotentials(shared, conf);

	///////////////////////////////////////////////////////////////////////////////
	// Forward
	///////////////////////////////////////////////////////////////////////////////

	computeLogPotential<T>(shared, 0, structure, conf);
	p = getLogNodePotential(shared, label, conf);

	storeSync(shared, p);
	clearPotentials(shared, conf);

	for(unsigned int t=1; t<L; t++){
		computeLogPotential<T>(shared, t, structure, conf);
		p = getLogNodePotential(shared, label, conf);

		maxValue = shared[0] + getLogEdgePotential(shared, 0, label, conf);
		maxIdx = 0;

		for(unsigned int i=0; i<blockDim.x; i++){
			q = shared[i] + getLogEdgePotential(shared, i, label, conf);

			if( q > maxValue ){
				maxValue = q;
				maxIdx = i;
			}
		}
		storeSync(shared, maxValue+p);

		setViterbiPath(structure, t-1, maxIdx, cConfig, conf);
		clearPotentials(shared, conf);
	}
	maxValue = shared[label];
	maxIdx = label;

	storeSync(shared, maxValue);

	if(label == 0){
		for(unsigned int i=0; i<blockDim.x; i++){
			q = shared[i];

			if( q > maxValue ){
				maxValue = q;
				maxIdx = i;
			}
		}
		cConfig->labels[conf->structMap[structure]+L-1] = maxIdx;
		for(int t=L-2; t>=0; t--){
			maxIdx = cConfig->viterbiPath[(conf->structMap[structure]+t)*blockDim.x+maxIdx];
			cConfig->labels[conf->structMap[structure]+t] = maxIdx;
		}
	}
}

template <class T>
void viterbiLCRFGPU(ForwardBackwardConfig<T>* conf, LinearModelConfig<T>* pConfig){

	const unsigned int threads   = pConfig->labelCount;
	const unsigned int smem      = (2*pConfig->labelCount+(pConfig->labelCount*pConfig->labelCount))*sizeof(T);

	unsigned int totalTasks = pConfig->structureCount;
	unsigned int tasks      = min(conf->maxBlocks, totalTasks);
	unsigned int start      = 0;

	dim3 viterbiBlock(threads, 1, 1);

	LinearModelConfig<T>* gpuPConfig;
	cudaMalloc((void**)&gpuPConfig, sizeof(LinearModelConfig<T>));
	cudaMemcpy(gpuPConfig, pConfig, sizeof(LinearModelConfig<T>), cudaMemcpyHostToDevice);

	ForwardBackwardConfig<T>* gpuConfig;
	cudaMalloc((void**)&gpuConfig, sizeof(ForwardBackwardConfig<T>));
	cudaMemcpy(gpuConfig, conf, sizeof(ForwardBackwardConfig<T>), cudaMemcpyHostToDevice);


	//cudaPrintfInit(64*1048576);

	while(totalTasks){
		dim3 viterbiGrid(tasks, 1, 1);

		viterbiLCRFKernel<T><<<viterbiGrid, viterbiBlock, smem>>>(start, gpuConfig, gpuPConfig);

		start += tasks;
		totalTasks -= tasks;
		tasks = min(conf->maxBlocks, totalTasks);
	}
	//cudaPrintfDisplay(stdout, true);
	//cudaPrintfEnd();

	cudaFree(gpuConfig);
	cudaFree(gpuPConfig);
}


template void viterbiLCRFGPU<float>(ForwardBackwardConfig<float>*, LinearModelConfig<float>*);
template void viterbiLCRFGPU<double>(ForwardBackwardConfig<double>*, LinearModelConfig<double>*);

#endif
