// Copyright (C) 2011 Nico Piatkowski <nico.piatkowski@tu-dortmund.de>, All rights reserved 
#ifndef ForwardBackward_H
#define ForwardBackward_H

#include <iostream>

#include <vector_types.h>
#include <cuda.h>
#include <cuda_runtime_api.h>

#include "AbstractLayer/InferenceEngine.h"
#include "LinearModel.h"
#include "ForwardBackwardConfig.h"

template <class T> void fbLCRFGPU(ForwardBackwardConfig<T>*, LinearModelConfig<T>*);
template <class T> void viterbiLCRFGPU(ForwardBackwardConfig<T>* conf, LinearModelConfig<T>* pConfig);
template <class T> void loadExpectations(T*, unsigned int, void*);

template <class T>
class ForwardBackward : public InferenceEngine<T>{

	private:
		LinearModel<T>* parameters;
		ForwardBackwardConfig<T>* conf;

		T* logProps;
		unsigned int logPropsAlloc;
		bool updateLogProps;


		ObservationSet<T>* lastSet;

		void setup(ObservationSet<T>* s){
			unsigned int size;

			size = s->getSize();
			cudaRealloc(&conf->logProps, conf->logPropsAlloc, size);

			size = (s->getNodeMapSize()-1)*parameters->getLabelCount();
			cudaRealloc(&conf->alpha, conf->alphaAlloc, size);

			size = (s->getNodeMapSize()-1)*parameters->getLabelCount();
			cudaRealloc(&conf->viterbiPath, conf->viterbiPathAlloc, size);

			size = (s->getNodeMapSize()-1);
			cudaRealloc(&conf->labels, conf->labelAlloc, size);
			//cudaMemcpy(conf->labels, s->getLabels(), size*sizeof(unsigned int), cudaMemcpyHostToDevice);

			cudaThreadSynchronize();
		};

	public:
		ForwardBackward(LinearModel<T>* p){
			parameters = p;
			logPropsAlloc = 0;
			logProps = NULL;
			lastSet  = NULL;
			updateLogProps = false;


			conf = (ForwardBackwardConfig<T>*) malloc( sizeof( ForwardBackwardConfig<T> ) );
			memset(conf, 0, sizeof(ForwardBackwardConfig<T>));

			cudaDeviceProp deviceProp;
			cudaGetDeviceProperties(&deviceProp, 0);
			conf->maxBlocks = deviceProp.maxGridSize[0];
		};

		~ForwardBackward(){

			if(conf->logProps) cudaFree(conf->logProps);
			if(conf->alpha) cudaFree(conf->alpha);
			if(conf->labels) cudaFree(conf->labels);
			if(conf->viterbiPath) cudaFree(conf->viterbiPath);

			free(conf);
		}

		T* getExpectations(){
			LinearModelConfig<T>* c = (LinearModelConfig<T>*)parameters->getConfig();

			loadExpectations(c->expect,getModel()->getDim(), parameters->getConfigGPU());

			return c->expect;
		};

		void computeMarginals(ObservationSet<T>* s){
			//if(s != lastSet){
				lastSet = s;
				setup(s);
			//}

			fbLCRFGPU(conf, (LinearModelConfig<T>*) parameters->getConfig());
	

			updateLogProps = true;
		};

		unsigned int* computeMaxPropAssignment(ObservationSet<T>* s){
			setup(s);
			viterbiLCRFGPU(conf, (LinearModelConfig<T>*) parameters->getConfig());
			unsigned int* mpl = (unsigned int*) malloc((s->getNodeMapSize()-1)*sizeof(unsigned int));
			cudaMemcpy(mpl, conf->labels, (s->getNodeMapSize()-1)*sizeof(unsigned int), cudaMemcpyDeviceToHost);
			return mpl;
		}

		void disableGradient(){
			
		}

		unsigned int getGPUMemUsage(){
			return 	conf->logPropsAlloc * sizeof(T)+
					conf->alphaAlloc * sizeof(T)+
					conf->viterbiPathAlloc * sizeof(unsigned int)+
					conf->labelAlloc * sizeof(unsigned int);
		}

		void setParameters(LinearModel<T>* p){
			parameters = p;
		};

		Model<T>* getModel(){
			return parameters;
		};

		T* getLogProps(){
			unsigned int size = conf->logPropsAlloc;

			if(updateLogProps){
				if(!logProps) logProps = (T*) malloc(size*sizeof(T));
				else if(size > logPropsAlloc){
					logProps = (T*) realloc(logProps, size*sizeof(T));
					logPropsAlloc = size;
				}

				cudaMemcpy(logProps, conf->logProps, size*sizeof(T), cudaMemcpyDeviceToHost);
				updateLogProps = false;
			}
			return logProps;
		};

		T* getLogPropsGPU(){ return conf->logProps; };
};
#endif
