// Copyright (C) 2011 Nico Piatkowski <nico.piatkowski@tu-dortmund.de>, All rights reserved

#ifndef LinearModel_CUH
#define LinearModel_CUH

#include "LinearModelConfig.h"

#define EPS (T)0.00001

enum PotentialMode { STD_NODE = 0, TRANSITION = 1 };

/*********************************************************************/

__device__ unsigned long baseEncode(unsigned int *a, unsigned int n, unsigned int b){
	unsigned long result = 0;
	unsigned int p = 1;

	for(unsigned int i=0; i<n; i++){
		result += a[i]*p;
		p *= b;
	}

	return result;
}

__device__ void baseDecode(volatile unsigned int *a, unsigned int n, unsigned int b, unsigned long m){
	unsigned long c = m;

	for(unsigned int i=0; i<n; i++){
		a[i] = c%b;
		c    = c/b;
	}
}

__device__ void baseDecode(unsigned int *a, unsigned int n, unsigned int b, unsigned long m){
	unsigned long c = m;

	for(unsigned int i=0; i<n; i++){
		a[i] = c%b;
		c    = c/b;
	}
}

__device__ void baseDecode(unsigned int& a0, unsigned int& a1, unsigned int b, unsigned long m){
	unsigned long c = m;

	//for(unsigned int i=0; i<n; i++){
		a0 = c%b;
		c    = c/b;
		a1 = c%b;
	//}
}

template <class T>
__device__ void expandIndex(unsigned int k, unsigned int &a, unsigned int &from, unsigned int &to, LinearModelConfig<T>* conf){
	a    = (k / (conf->labelCount+(conf->labelCount*conf->labelCount)));
	to   = (k - a*(conf->labelCount+(conf->labelCount*conf->labelCount))) / (1+conf->labelCount);
	from = k - a*(conf->labelCount+(conf->labelCount*conf->labelCount)) - to*(1+conf->labelCount);
}


template <class T>
__device__ void clearPotentials(volatile T* sdata, LinearModelConfig<T>* conf){
	for(unsigned int i = threadIdx.x; i < (conf->labelCount*conf->labelCount); i+=blockDim.x){
		sdata[conf->labelCount+i] = (T)0.0;
	}
	for(unsigned int i = threadIdx.x; i < conf->labelCount; i+=blockDim.x){
		sdata[conf->labelCount+(conf->labelCount*conf->labelCount)+i] = (T)0.0;
	}
	__syncthreads();
}

template <class T>
__device__ T getEdgePotential(volatile T* sdata, unsigned int from, unsigned int to, LinearModelConfig<T>* conf){
	return exp(sdata[conf->labelCount+(from*conf->labelCount)+to]);
}

template <class T>
__device__ T getLogEdgePotential(volatile T* sdata, unsigned int from, unsigned int to, LinearModelConfig<T>* conf){
	return sdata[conf->labelCount+(from*conf->labelCount)+to];
}

template <class T>
__device__ void setLogNodePotential(volatile T* sdata, unsigned int j, T val, LinearModelConfig<T>* conf){
	__syncthreads();
	sdata[conf->labelCount+(conf->labelCount*conf->labelCount)+j] = val;
	__syncthreads();
}

template <class T>
__device__ void addLogNodePotential(volatile T* sdata, unsigned int j, T val, LinearModelConfig<T>* conf){
	__syncthreads();
	sdata[conf->labelCount+(conf->labelCount*conf->labelCount)+j] += val;
	__syncthreads();
}

template <class T>
__device__ void addLogEdgePotential(volatile T* sdata, unsigned int i, unsigned int j, T val, LinearModelConfig<T>* conf){
	__syncthreads();
	sdata[conf->labelCount+(i*conf->labelCount)+j] += val;
	__syncthreads();
}

template <class T>
__device__ T getLogNodePotential(volatile T* sdata, unsigned int j, LinearModelConfig<T>* conf){
	return sdata[conf->labelCount+(conf->labelCount*conf->labelCount)+j];
}


template <class T>
__device__ void setNodeMarginal(unsigned int structure, int t, T value, LinearModelConfig<T>* conf){
	conf->nMarg[(conf->structMap[structure] + t)*conf->labelCount+threadIdx.x] = value;
}

template <class T>
__device__ T getNodeMarginal(unsigned int idx, unsigned int y, LinearModelConfig<T>* conf){
	return conf->nMarg[(idx*(conf->labelCount))+y];
}

template <class T>
__device__ void setEdgeMarginal(unsigned int structure, int t, unsigned int from, T value, LinearModelConfig<T>* conf){
	conf->eMarg[(conf->structMap[structure] + t)*(conf->labelCount*conf->labelCount)+(from*conf->labelCount)+threadIdx.x] = value;
}

template <class T>
__device__ T getEdgeMarginal(unsigned int idx, unsigned int from, unsigned int to, LinearModelConfig<T>* conf){
	return conf->eMarg[idx*(conf->labelCount*conf->labelCount)+(from*conf->labelCount)+to];
}


template <class T>
__device__ void computeLogPotential(volatile T* sdata, unsigned int i, unsigned int structure, LinearModelConfig<T>* conf){

	int l, u;
	unsigned int a, from, to;

	T val = (T)0.0;

	for(unsigned int j = conf->nodeMap[conf->structMap[structure]+i]; j < conf->nodeMap[conf->structMap[structure]+1+i]; j++){

		l = conf->attStateMap[conf->obs[j]].x + threadIdx.x;
		u = conf->attStateMap[conf->obs[j]].y;

		if(l < u && l!=-1 && u != -1){
			val = conf->values[j];

			a = conf->parMap[l].a; baseDecode(from, to, 67, conf->parMap[l].j); to--;
			
			if(a == conf->obs[j]){
				if(!from){
					addLogNodePotential(sdata, to, conf->parameter[l] * val, conf);
				}
				else{
					from--;
					addLogEdgePotential(sdata, from, to, conf->parameter[l] * val, conf);
				}
			}

			l += blockDim.x;
		}
	}

	__syncthreads();
}

template <class T>
__device__ void computeLogNodePotential(volatile T* sdata, unsigned int i, unsigned int structure, LinearModelConfig<T>* conf){

	int l, u;
	unsigned int a, from, to;

	T val = (T)0.0;

	for(unsigned int j = conf->nodeMap[conf->structMap[structure]+i]; j < conf->nodeMap[conf->structMap[structure]+1+i]; j++){

		l = conf->attStateMap[conf->obs[j]].x + threadIdx.x;
		u = conf->attStateMap[conf->obs[j]].y;

		if(l < u && l!=-1 && u != -1){
			val = conf->values[j];

			a = conf->parMap[l].a; baseDecode(from, to, 67, conf->parMap[l].j); to--;
			
			if(a == conf->obs[j]){
				if(!from){
					addLogNodePotential(sdata, to, conf->parameter[l] * val, conf);
				}
				else{
					u = -1;
				}
			}

			l += blockDim.x;
		}
	}

	__syncthreads();
}

template <class T, PotentialMode M>
__device__ T sumMarginals(unsigned int a, int l, int u, unsigned int f, unsigned int to, LinearModelConfig<T>* conf){

	unsigned int obs, from;
	T res = (T)0.0, p = (T)0.0;

	if(M == TRANSITION) from = f - TRANSITION;

	if(l!=-1 && u != -1) for(int i = l; i < u; i++){
		obs = conf->attributeObserved[i].x;
		p = (T)0.0;

		if(M == STD_NODE   ) p = getNodeMarginal<T>(obs,              to, conf);
		if(M == TRANSITION ) p = getEdgeMarginal<T>(obs, from,        to, conf);

		res += p;
	}

	return res;
}

template <class T>
__device__ T getModelExpect(unsigned int pId, void* vconf){

		LinearModelConfig<T>* conf = (LinearModelConfig<T>*)vconf;

		int l=0, u=0;
		unsigned int att=0, nodeType=0, to=0;
		T temp = (T)0.0;

		att = conf->parMap[pId].a; baseDecode(nodeType, to, 67, conf->parMap[pId].j); to--;

		l = conf->attributeObservedIdx[att].x;
		u = conf->attributeObservedIdx[att].y;

		switch(nodeType){
			case STD_NODE:   { temp = sumMarginals<T, STD_NODE>   (att, l, u, nodeType, to, conf); break; }
			default:         { temp = sumMarginals<T, TRANSITION> (att, l, u, nodeType, to, conf); break; }
		}

		return temp;
}

template <class T>
__global__ void loadExpectationsKernel(T* expect, unsigned int dim, void* conf){

	unsigned int threadID = blockIdx.x*blockDim.x+threadIdx.x;

	while(threadID < dim){

		expect[threadID] = getModelExpect<T>(threadID, conf);

		threadID += blockDim.x*gridDim.x;
	}
}

template <class T>
void loadExpectations(T* expect, unsigned int dim, void* conf){

	dim3 grid(240, 1, 1);
	dim3 block(32, 1, 1);

	loadExpectationsKernel<T><<<grid, block>>>(expect, dim, conf);
	cudaFree(conf);
}

template void loadExpectations<float>(float*, unsigned int,void*);
template void loadExpectations<double>(double*, unsigned int,void*);

template <class T>
__device__ void storeSync(volatile T* sdata, T val){
	__syncthreads();
	sdata[threadIdx.x] = val;
	__syncthreads();
}

template <class T>
__device__ T sumAndClear(volatile T* sdata){
	T q = (T)0.0;
	for(unsigned int i = 0; i < blockDim.x; i++){
		q += sdata[i];
		sdata[blockDim.x+(i*blockDim.x+threadIdx.x)] = (T)0.0;
	}
	return (fabsf(q) > (T)0+EPS) ? q : (T)1.0;
}

/*********************************************************************/
#endif
