// Copyright (C) 2011 Nico Piatkowski <nico.piatkowski@tu-dortmund.de>, All rights reserved 
#ifndef LinearModel_H
#define LinearModel_H

#include <malloc.h>
#include <map>
#include <set>
#include <algorithm>
#include <fstream>

#include <vector_types.h>
#include <cuda.h>
#include <cuda_runtime_api.h>

#include "AbstractLayer/Model.h"
#include "misc/MemUtil.h"
#include "LinearModelConfig.h"
# include <ext/hash_map>

template <class T>
class LinearModel : public Model<T>{
	private:
	
		LinearModelConfig<T>* conf;

		__gnu_cxx::hash_map<ParameterDescriptor, int>* parameterCounts;
		__gnu_cxx::hash_map<unsigned int, std::vector<int2>* >* attObservationMap;

		std::map<ParameterDescriptor, T>** fastpVals;


		int2* attributeObserved;

		int2* attributeObservedIdx;
		int2* attStateMap;

		ParameterDescriptor* parameterMap;
		unsigned int* counts;

		unsigned int parametersAlloc;
		unsigned int parameterMapAlloc;
		unsigned int countAlloc;

		unsigned int attStateMapAlloc;

		unsigned int attributeObservedAlloc;
		unsigned int attributeObservedIdxAlloc;

		unsigned int observedCount;
		unsigned int attStateSize;

		unsigned int tlength;

		unsigned int time;

		// 0  1       2     3
		// y, source, sink, transitions..
		const unsigned int transitionOffset;

		unsigned int labels;
		unsigned int lastAttributeIndex;
		unsigned int attTableSize;

		bool* seenAtts;

		using Model<T>::parameters;
		using Model<T>::length;
		using Model<T>::graph;
		using Model<T>::init_stepsize;
		using Model<T>::init_weight;

		ObservationSet<T>* lastO;

		T* gbuf;
		T* h_gbuf;

		void setup(ObservationSet<T>* s){

			lastO = s;

			conf->structureCount = s->getSize();
			conf->labelCount     = getLabelCount();

			unsigned int nodeMarginalsSize = (s->getNodeMapSize()-1)*conf->labelCount;
			unsigned int edgeMarginalsSize = (s->getNodeMapSize()-1)*conf->labelCount*conf->labelCount;

			cudaRealloc<T>(&conf->nMarg, conf->nodeMarginalsAlloc, nodeMarginalsSize);

			cudaRealloc<T>(&conf->eMarg, conf->edgeMarginalsAlloc, edgeMarginalsSize);
			cudaThreadSynchronize();

			clearGPU<T>(conf->nMarg, conf->nodeMarginalsAlloc);
			clearGPU<T>(conf->eMarg, conf->edgeMarginalsAlloc);

			unsigned int size = length;

			cudaRealloc(&conf->counts, conf->countAlloc, size);
			cudaMemcpy(conf->counts, counts, size*sizeof(unsigned int), cudaMemcpyHostToDevice);

			size = length;
			cudaRealloc<T>(&conf->parameter, conf->parameterAlloc, size);
			cudaMemcpy(conf->parameter, parameters, size*sizeof(T), cudaMemcpyHostToDevice);


			size = length;
			cudaRealloc(&conf->parMap, conf->parameterMapAlloc, size);
			cudaMemcpy(conf->parMap, parameterMap, size*sizeof(ParameterDescriptor), cudaMemcpyHostToDevice);

			size = length;
			cudaRealloc(&conf->expect, conf->expectAlloc, size);
			clearGPU<T>(conf->expect,	conf->expectAlloc);

			size = s->getSize();
			cudaRealloc(&conf->structureSizes, conf->structureSizesAlloc, size);
			cudaMemcpy(conf->structureSizes, s->getStructureSizes(), size*sizeof(unsigned int), cudaMemcpyHostToDevice);

			size = s->getStructureMapSize();
			cudaRealloc(&conf->structMap, conf->structMapAlloc, size);
			cudaMemcpy(conf->structMap, s->getStructureMap(), size*sizeof(unsigned int), cudaMemcpyHostToDevice);

			size = s->getNodeMapSize();
			cudaRealloc(&conf->nodeMap, conf->nodeMapAlloc, size);
			cudaMemcpy(conf->nodeMap, s->getNodeMap(), size*sizeof(unsigned int), cudaMemcpyHostToDevice);

			size = s->getObservationSize();
			cudaRealloc(&conf->obs, conf->observationAlloc, size);
			cudaMemcpy(conf->obs, s->getObservations(), size*sizeof(unsigned int), cudaMemcpyHostToDevice);

			size = s->getObservationSize();
			cudaRealloc(&conf->values, conf->valuesAlloc, size);
			cudaMemcpy(conf->values, s->getValues(), size*sizeof(T), cudaMemcpyHostToDevice);

			size = getAttributeLabelParameterMapSize();
			cudaRealloc(&conf->attStateMap, conf->attStateMapAlloc, size);
			cudaMemcpy(conf->attStateMap, getAttributeLabelParameterMap(), size*sizeof(int2), cudaMemcpyHostToDevice);

			size = observedCount;
			cudaRealloc(&conf->attributeObserved, conf->attributeObservedAlloc, size);
			cudaMemcpy(conf->attributeObserved, attributeObserved, size*sizeof(int2), cudaMemcpyHostToDevice);

			size = lastAttributeIndex+2;
			cudaRealloc(&conf->attributeObservedIdx, conf->attributeObservedIdxAlloc, size);
			cudaMemcpy(conf->attributeObservedIdx, attributeObservedIdx, size*sizeof(int2), cudaMemcpyHostToDevice);
		}

	public:
		LinearModel(AbstractGraph* g, unsigned int l) : transitionOffset(1), Model<T>(){
			graph	= g;
			labels	= l;
			time	= 1;

			parameterCounts   = new __gnu_cxx::hash_map<ParameterDescriptor, int>();
			attObservationMap = new __gnu_cxx::hash_map<unsigned int, std::vector<int2>* >();

			//subBlockSize = transitionOffset + labels;
			//blockSize	 = subBlockSize * labels;

			conf = (LinearModelConfig<T>*) myMalloc( sizeof(LinearModelConfig<T>) );
			memset(conf, 0, sizeof(LinearModelConfig<T>));

			attTableSize = 50000000;

			seenAtts = new bool[attTableSize];
			fastpVals = new std::map<ParameterDescriptor, T>*[attTableSize];

			memset(fastpVals, 0, sizeof(std::map<ParameterDescriptor, T>*)*attTableSize);

			memset(seenAtts, 0, sizeof(bool)*attTableSize);
			//conf->blockSize    = blockSize;
			//conf->subBlockSize = subBlockSize;

			observedCount = 0;


			unsigned int preAlloc = 1;

			parameters   = (T*)myMalloc(preAlloc*sizeof(T));
			parameterMap = (ParameterDescriptor*)myMalloc(preAlloc*sizeof(ParameterDescriptor));
			counts       = (unsigned int*)myMalloc(preAlloc*sizeof(unsigned int));

			attributeObserved    = (int2*)myMalloc(preAlloc*sizeof(int2));
			attributeObservedIdx = (int2*)myMalloc(preAlloc*sizeof(int2));

			attStateMap  = (int2*)myMalloc(preAlloc*sizeof(int2));

			countAlloc        = preAlloc;
			parametersAlloc   = preAlloc;
			parameterMapAlloc = preAlloc;

			attributeObservedAlloc    = preAlloc;
			attributeObservedIdxAlloc = preAlloc;

			attStateMapAlloc  = preAlloc;

			length = 0;
			tlength = 0;

		};


		std::map<ParameterDescriptor, T>* getParameterValues(){
			return NULL;
		}

		void saveToFile(std::string s){
			lastO->getClassDict()->addToFile(s);
			lastO->getAttDict()->addToFile(s);

			typename std::map<ParameterDescriptor, T>::const_iterator jj;

			ParameterDescriptor* p = (ParameterDescriptor*)malloc(sizeof(ParameterDescriptor));

			std::fstream ff(s.c_str());
			ff.seekp(0, ios_base::end);

			for(unsigned int i = 0; i<attTableSize; i++){
				if(fastpVals[i]){
					for(jj = fastpVals[i]->begin(); jj != fastpVals[i]->end(); jj++){
						*p = jj->first;
						ff << p->a << '\t' << p->i << '\t' << p->j << '\t' << p->jIdx << '\t' << p->n << '\t' << p->k << '\t' << jj->second << std::endl;
					}
				}
			}
			ff << "[ENDPAR]" << std::endl;
			ff.close();
			free(p);
			//std::cout << "Last attribute Index = " << lastAttributeIndex << std::endl;
		}

		void loadFromFile(std::string s){
			unsigned int k = 0;
			std::string line;

			std::fstream fff(s.c_str());
			std::getline(fff, line);
			std::getline(fff, line);
			k = fff.tellg();
			fff.close();

			k = lastO->getClassDict()->loadFromFile(s, k);
			k = lastO->getAttDict()->loadFromFile(s, k);

			std::fstream ff(s.c_str());
			ff.seekg(k);
			std::string mark("[ENDPAR]");

			std::getline(ff, line);
			while(line.compare(mark)!=0){

				strtokenizer* tok = new strtokenizer(&line, " \t");

				ParameterDescriptor* p = (ParameterDescriptor*)malloc(sizeof(ParameterDescriptor));

				stringstream sstra(*(tok->token(0)));
				unsigned int a;
				sstra >> a;
				p->a = a;

				stringstream sstri(*(tok->token(1)));
				unsigned long i;
				sstri >> i;
				p->i = i;

				stringstream sstrj(*(tok->token(2)));
				unsigned long j;
				sstrj >> j;
				p->j = j;

				stringstream sstrji(*(tok->token(3)));
				unsigned int ji;
				sstrj >> ji;
				p->jIdx = ji;

				stringstream sstrn(*(tok->token(4)));
				unsigned int n;
				sstrn >> n;
				p->n = n;

				stringstream sstrkk(*(tok->token(5)));
				unsigned int kk;
				sstrkk >> kk;
				p->k = kk;

				stringstream sstrpar(*(tok->token(6)));
				T par;
				sstrpar >> par;

				if(!fastpVals[a]) fastpVals[a] = new std::map<ParameterDescriptor, T>();

				(*(fastpVals[a]))[*p] = par;
				tlength++;
				length++;

				lastAttributeIndex = a;
				std::getline(ff, line);
			}
			attStateMap = (int2*)myRealloc(attStateMap, (lastAttributeIndex+2)*sizeof(int2));
			attStateMapAlloc = (lastAttributeIndex+2);

			ff.close();
		}

		void* getConfig(){ return conf; }
		void* getConfigGPU(){

			LinearModelConfig<T>* gpuPConfig;
			cudaMalloc((void**)&gpuPConfig, sizeof(LinearModelConfig<T>));
			cudaMemcpy(gpuPConfig, conf, sizeof(LinearModelConfig<T>), cudaMemcpyHostToDevice);

			return gpuPConfig;
		}

		~LinearModel(){
			if(parameters)   cudaFreeHost(parameters);
			if(parameterMap) cudaFreeHost(parameterMap);
			if(counts)       cudaFreeHost(counts);

			if(attributeObservedIdx) cudaFreeHost(attributeObservedIdx);
			if(attributeObserved) cudaFreeHost(attributeObserved);

			if(attStateMap)  cudaFreeHost(attStateMap);

			if(conf->parameter) cudaFree(conf->parameter);
			if(conf->nMarg) cudaFree(conf->nMarg);
			if(conf->eMarg) cudaFree(conf->eMarg);
			if(conf->structureSizes) cudaFree(conf->structureSizes);
			if(conf->parMap) cudaFree(conf->parMap);
			if(conf->structMap) cudaFree(conf->structMap);
			if(conf->nodeMap) cudaFree(conf->nodeMap);
			if(conf->attStateMap) cudaFree(conf->attStateMap);
			if(conf->obs) cudaFree(conf->obs);
			if(conf->counts) cudaFree(conf->counts);
			if(conf->expect) cudaFree(conf->expect);

			if(conf->attributeObservedIdx) cudaFree(conf->attributeObservedIdx);
			if(conf->attributeObserved) cudaFree(conf->attributeObserved);


			free(conf);


			for(__gnu_cxx::hash_map<unsigned int, std::vector<int2>* >::const_iterator ii = attObservationMap->begin(); ii != attObservationMap->end(); ii++)
				delete ii->second;
		}

		unsigned int getGPUMemUsage(){
			return  conf->structureSizesAlloc * sizeof(unsigned int)+
					conf->structMapAlloc * sizeof(unsigned int)+
					conf->nodeMarginalsAlloc * sizeof(T)+
					conf->edgeMarginalsAlloc * sizeof(T)+
					conf->parameterAlloc * sizeof(T)+
					conf->parameterMapAlloc * sizeof(ParameterDescriptor)+
					conf->nodeMapAlloc * sizeof(unsigned int)+
					conf->observationAlloc * sizeof(unsigned int)+
					conf->attStateMapAlloc * sizeof(int2)+
					conf->attributeObservedIdxAlloc * sizeof(int2)+
					conf->attributeObservedAlloc * sizeof(unsigned int)+
					conf->countAlloc * sizeof(unsigned int)+
					conf->valuesAlloc * sizeof(T);
		}

		T* getParametersGPU(){
			return conf->parameter;
		}

		unsigned int* getCounts(){ return counts; };
		unsigned int* getCountsGPU(){ return conf->counts; };

		unsigned int encodeIndex(unsigned int* enc, unsigned int c = 2){

			unsigned int k = 0;
			unsigned int j = transitionOffset+labels;

			// attribut * sum{i = 0 to c}{labels^i} + 

			for(unsigned int i=0; i<c; i++){
				k += enc[i] * j;
				j *= labels;
			}

			return k;
		}

/*

	j = 1;
	j = 1+labels;
	j = labels * (1+labels)
	j = labels * labels * (1+labels)

(a,y),(a,y,y_0),(a,y,y_2),(a,y,y_7), (a,y,y_7)

*/

		void expandIndex(unsigned int idx, unsigned int &i, unsigned int &from, unsigned int &to){
			i    = idx / (labels+(labels*labels));
			to   = (idx - i*(labels+(labels*labels))) / (1+labels);
			from = idx - i*(labels+(labels*labels)) - to*(1+labels);
		};


		void addParameter(ParameterDescriptor cd){
			(*parameterCounts)[cd]++;

			std::map<ParameterDescriptor, T>* pMap;

			if(fastpVals[cd.a] != NULL){
				pMap = fastpVals[cd.a];

			}else{
				pMap = new std::map<ParameterDescriptor, T>();

				fastpVals[cd.a] = pMap;

			}

			typename std::map<ParameterDescriptor, T>::const_iterator w = pMap->find(cd);

			if(w == pMap->end()){
				(*pMap)[cd] = init_weight;
				tlength++;
				length++;
			}
		};

		int addAttribute(unsigned int a, unsigned int set_nodeId, unsigned int i, unsigned int nodeCount, unsigned int dId, unsigned int* nodeLabel, std::map<unsigned int, std::vector<ParameterDescriptor>* >* waitingNodes, bool update){
			observedCount++;
			if(attObservationMap->find(a) == attObservationMap->end()) (*attObservationMap)[a] = new std::vector<int2>();
			int2 tt;
			tt.x = set_nodeId;
			(*attObservationMap)[a]->push_back(tt);

			unsigned int *e, *m, other;

			ParameterDescriptor*  cd = (ParameterDescriptor*)malloc(sizeof(ParameterDescriptor));
			memset(cd, 0, sizeof(ParameterDescriptor));

			cd->a = a;
			cd->i = 0;
			cd->j = AbstractGraph::baseEncode((unsigned int)0, nodeLabel[i]+1, 67);
			cd->n = 2;

			graph->setNodeCount(nodeCount);
			e = graph->getEdges();
			m = graph->getEdgeMap();

			if(update ) addParameter(*cd);

			// for all j: (j,i) \in E
			for(unsigned int k=m[i];  update && k<m[i+1]; k++){

				other = e[k];

				if(other <= i){
					cd->j = AbstractGraph::baseEncode(nodeLabel[other]+1, nodeLabel[i]+1, 67);
					addParameter(*cd);
				}
			}

			free(cd);

			return 0;
		}

		void clear(){
			observedCount = 0;
			length = 0;

			for(__gnu_cxx::hash_map<unsigned int, std::vector<int2>* >::const_iterator ii = attObservationMap->begin(); ii != attObservationMap->end(); ii++)
				delete ii->second;

			attObservationMap->clear();
			parameterCounts->clear();
		}

		void store(ObservationSet<T>* s){

			s->setPCounts(parameterCounts);
			s->setAOMap(attObservationMap);
			s->setOCount(observedCount);
			s->setLen(length);

			observedCount = 0;
			length = 0;

			parameterCounts   = new __gnu_cxx::hash_map<ParameterDescriptor, int>();
			attObservationMap = new __gnu_cxx::hash_map<unsigned int, std::vector<int2>* >();
		}

		void init(ObservationSet<T>* s){

			if(s->getPCounts() != NULL){
				parameterCounts = s->getPCounts();
				attObservationMap = s->getAOMap();
				observedCount = s->getOCount();
				length = s->getLen();
			}

			unsigned int att = 0, to = 0, from = 0, pIdx = 0, lf = 0;
			lastAttributeIndex = s->getLastAttributeIndex();
			//std::cout << std::endl << lastAttributeIndex << std::endl;

			if(observedCount > attributeObservedAlloc){
				attributeObserved = (int2*)myRealloc(attributeObserved, observedCount*sizeof(int2));
				attributeObservedAlloc = observedCount;
			}

			int2 tempI2;

			tempI2.x = -1;
			tempI2.y = -1;


			unsigned int size = lastAttributeIndex+2;
			if(size > attributeObservedIdxAlloc){
				attributeObservedIdx = (int2*)myRealloc(attributeObservedIdx, size*sizeof(int2));
				attributeObservedIdxAlloc = size;
			}

			std::fill(attributeObservedIdx, attributeObservedIdx+size, tempI2);

			unsigned int counter = 0;
			int lastaoi = -1;

			for(__gnu_cxx::hash_map<unsigned int, std::vector<int2>* >::const_iterator ii = attObservationMap->begin(); ii != attObservationMap->end(); ii++){

				attributeObservedIdx[ii->first].x = counter;
				if(lastaoi != -1) attributeObservedIdx[lastaoi].y = counter;
				lastaoi = ii->first;

				for(std::vector<int2>::const_iterator jj = ii->second->begin(); jj != ii->second->end(); jj++){
					attributeObserved[counter++] = *jj;
				}
			}
			//attributeObservedIdx[lastAttributeIndex+1] = counter;

			/////////////////////////////////////////////////////////////
			// PARAMETER, COUNT, MAP
			/////////////////////////////////////////////////////////////

			//std::cout << "Length: " << length << std::endl;

			if(tlength > parametersAlloc){
				parameters = (T*)myRealloc(parameters, tlength*sizeof(T));
				parametersAlloc = tlength;
			}

			if(tlength > parameterMapAlloc){
				parameterMap = (ParameterDescriptor*)myRealloc(parameterMap, tlength*sizeof(ParameterDescriptor));
				parameterMapAlloc = tlength;
			}

			if(tlength > countAlloc){
				counts = (unsigned int*)myRealloc(counts, tlength*sizeof(unsigned int));
				countAlloc = tlength;
			}

			/////////////////////////////////////////////////////////////

			size = (lastAttributeIndex+2);
			if(size > attStateMapAlloc){
				attStateMap = (int2*)myRealloc(attStateMap, size*sizeof(int2));
				attStateMapAlloc = size;
			}
			std::fill(attStateMap, attStateMap+size, tempI2);
			attStateSize = size;

			/////////////////////////////////////////////////////////////

			typename __gnu_cxx::hash_map<unsigned int, std::vector<int2>* >::const_iterator ii = attObservationMap->begin();

			typename std::map<ParameterDescriptor, T>::const_iterator jj;

			ParameterDescriptor* parameterId = (ParameterDescriptor*)malloc(sizeof(ParameterDescriptor));

			int lastatt = -1;

			for(; ii != attObservationMap->end(); ii++){
	
				for(jj = fastpVals[ii->first]->begin(); jj != fastpVals[ii->first]->end(); jj++){
					*parameterId = jj->first;
			
					att = parameterId->a; AbstractGraph::baseDecode(from, to, 67, parameterId->j); to--;

					if(lastatt != att){
						//std::cout << att << std::endl;
						attStateMap[att].x = pIdx; // auffuellen
						if(lastatt>=0)attStateMap[lastatt].y = pIdx; // auffuellen
					}

					lastatt = att;

					if(parameterCounts->find(*parameterId)==parameterCounts->end()) counts[pIdx] = 0;
					else counts[pIdx] = (*parameterCounts)[*parameterId];

					//std::cout << '\t' << pIdx << '\t' << att << '\t' << from << '\t' << to << std::endl;

					parameters[pIdx] = jj->second;
					parameterMap[pIdx] = *parameterId;

					pIdx++;
				}

			}

			free(parameterId);
			length = pIdx;
			setup(s);
		};

		void synchronizeParameters(){
			cudaThreadSynchronize();
			cudaMemcpy(parameters, conf->parameter, length*sizeof(T), cudaMemcpyDeviceToHost);
			
			for(unsigned int i=0; i<length; i++){
				unsigned int a = parameterMap[i].a;
				(*fastpVals[a])[parameterMap[i]] = parameters[i];
			}
			cudaThreadSynchronize();
		};

		unsigned int getTotalParameterCount(){
			return tlength;
		};


		unsigned int getAttributeLabelParameterMapSize(){
			return attStateSize;
		};

		unsigned int getLabelCount(){
			return labels;
		};

		int2* getAttributeLabelParameterMap(){ return attStateMap; };
		ParameterDescriptor* getParameterMap(){ return parameterMap; };
};
#endif
