// Copyright (C) 2011 Nico Piatkowski <nico.piatkowski@tu-dortmund.de>, All rights reserved 

#include <time.h>
#include <iostream>
#include <iomanip>
#include <sstream>
#include <string>
#include <boost/program_options.hpp>

#include "crfgpu.h"

enum Engine   { E_FB };
enum Datatype { D_FLOAT, D_DOUBLE };
enum OEngine  { O_SGD, O_ASGD };
enum GraphType{ G_CHAIN };

namespace po = boost::program_options;

struct Options{
	OEngine  o;
	Datatype d;
	Engine   e;
	GraphType g;

	float	 stepsize;
	float	 lambda;
	float	 mu;
	float	 weight;

	unsigned int labelcount;
	unsigned int batchsize;
	unsigned int iterations;
	unsigned int nodes;

	unsigned int blockBase;
	unsigned int threadBase;

	unsigned int blockBaseBP;
	unsigned int threadBaseBP;
	unsigned int threadBaseBP2;

	bool showPRF;
	bool clearScreen;
	bool showProgress;
	bool writeLabels;
	bool iobLabels;
	bool tieParameters;
	bool hasModel;
	bool test;
	bool predict;

	std::string label_file;
	std::string test_file;
	std::string model_file;
};

int parseCMD(unsigned int, char**, Options&);
template <class T> void evaluate(Options&, unsigned int, unsigned int, unsigned int, float, Model<T>*, InferenceEngine<T>*, ObservationSet<T>*, std::vector<ObservationSet<T>* > &);
template <class T> void readTestset(Options&, unsigned int&, Model<T>*, ObservationSet<T>*, std::vector<ObservationSet<T>* >&);
template <class T> void run(Options&);

/////////////////////////////////////////////////////////////////////////////////////////////////////////

int main(int argc, char** argv){

	Options crfopts;
	if(parseCMD(argc, argv, crfopts)) return 1;

	if(crfopts.o == O_SGD || crfopts.o == O_ASGD){
		if(crfopts.d == D_FLOAT){
			run<float>(crfopts);
		}else{
			run<double>(crfopts);
		}
	}

	return 0;
}

template <class T> void run(Options& crfopts){
	srand(clock());
  	clock_t start = clock(), a;
	unsigned long __start = clock();

	std::vector<ObservationSet<T>* > batches;
	std::vector<ObservationSet<T>* > testsets;

	float trainingTime = 0.0f;
	unsigned int total = 0;
	unsigned int tests = 0;
	unsigned int bytes = 0;
	unsigned int counter = 0;

	/////////////////////////////////////////////////////////////////////////////////////////////////////////
	// INIT FRAMEWORK
	/////////////////////////////////////////////////////////////////////////////////////////////////////////

	if(crfopts.hasModel && crfopts.predict){
		std::string s;
		std::fstream ff(crfopts.model_file.c_str());
		std::getline(ff, s);
		std::stringstream inferS(s);
		unsigned int mm;
		inferS >> mm;
		if(mm==0) crfopts.e = E_FB;
		else assert(false);
		std::getline(ff, s);
		unsigned int mg;
		std::stringstream graphS(s);
		graphS >> mg;
		if(mm==0) crfopts.g = G_CHAIN;
		else assert(false);
		ff.close();
	}

	AbstractGraph*		G = new LinearChain(crfopts.nodes);
	Model<T>*		M = new LinearModel<T>(G, crfopts.labelcount);
	InferenceEngine<T>*	P = new ForwardBackward<T>((LinearModel<T>*)M);

	M->setInitStepsize(crfopts.stepsize);
	M->setInitWeight(crfopts.weight);

	ObservationSet<T>*	B = new ObservationSet<T>(M), *nextB;
	LogLikelihood<T>*	L = new LogLikelihood<T>(P, B, crfopts.blockBase, crfopts.threadBase);
	Optimizer<T>*		OPT = new SGD<T>(crfopts.stepsize, L, crfopts.blockBase, crfopts.threadBase);

	if(crfopts.hasModel && crfopts.predict){
		M->init(B);
		std::cout << "LOADING MODEL\t" << std::flush;
		M->loadFromFile(crfopts.model_file);
		std::cout << "DONE" << std::endl;
	}

	/////////////////////////////////////////////////////////////////////////////////////////////////////////
	// TRAINING - 1st ITERATION (FROM DISK)
	/////////////////////////////////////////////////////////////////////////////////////////////////////////

	while(!std::cin.eof() && !crfopts.predict){

		a = clock();
		bytes += B->readFromStream(&std::cin, crfopts.batchsize);
		assert(B->getLabelCount() <= crfopts.labelcount);

		if( B->getSize() ){
			total += B->getSize();

			M->init(B);
			OPT->iterate();
			if(crfopts.showProgress) std::cout << "\rTRAINING\t" << total << std::flush;

			M->store(B);
			batches.push_back(B);
			nextB = new ObservationSet<T>(B->getAttDict(), B->getClassDict(), M, true);
			B = nextB;
			L->setObservations(B);
		}
		trainingTime += (float)(clock()-a)/(float)CLOCKS_PER_SEC;
	}
	std::cout << std::endl;
	/////////////////////////////////////////////////////////////////////////////////////////////////////////
	// READ TEST-SET (FROM DISK)
	/////////////////////////////////////////////////////////////////////////////////////////////////////////

	if(crfopts.test) readTestset(crfopts, tests, M, B, testsets);

	if(crfopts.test){ if(crfopts.showPRF || 1==crfopts.iterations) evaluate(crfopts, 0, total, tests, trainingTime, M, P, B, testsets);}
	else if(crfopts.showPRF || 1==crfopts.iterations) evaluate(crfopts, 0, total, total, trainingTime, M, P, B, batches);

	/////////////////////////////////////////////////////////////////////////////////////////////////////////
	// TRAINING (FROM RAM)
	/////////////////////////////////////////////////////////////////////////////////////////////////////////

	for(unsigned int i=1; i<crfopts.iterations && !crfopts.predict; i++){

		if(crfopts.o == O_ASGD) OPT->setStepsize(crfopts.stepsize/(float)(i+1));

		a = clock();
		typename std::vector<ObservationSet<T>* >::const_iterator bI = batches.begin();

		counter = 0;

		for(; bI != batches.end(); bI++){
			L->setObservations(*bI);
			counter += (*bI)->getSize();
			M->init(*bI);
			OPT->iterate();
			if(crfopts.showProgress) std::cout << "\rTRAINING\t" << counter << "/" << total << "  \t(" << (i+1) << ")"<< std::flush;
			M->store(*bI);
		}
		trainingTime += (float)(clock()-a)/(float)CLOCKS_PER_SEC;

		if(crfopts.test){ if(crfopts.showPRF || i+1==crfopts.iterations) evaluate(crfopts, i, total, tests, trainingTime, M, P, B, testsets);}
		else if(crfopts.showPRF || i+1==crfopts.iterations) evaluate(crfopts, i, total, total, trainingTime, M, P, B, batches);
	}

	if(crfopts.hasModel && !crfopts.predict){
		if(crfopts.showProgress) std::cout << "WRITING MODEL\t" << std::flush;
		std::ofstream ff(crfopts.model_file.c_str());
		ff << crfopts.e << std::endl;
		ff << crfopts.g << std::endl;
		ff.close();
		M->saveToFile(crfopts.model_file);
		std::cout << "DONE" << std::endl;
	}

	/////////////////////////////////////////////////////////////////////////////////////////////////////////
	// EXIT
	/////////////////////////////////////////////////////////////////////////////////////////////////////////
	if(crfopts.showProgress && !crfopts.predict){
		std::cout << "LABELS\t\t" << B->getLabelCount()<< std::endl;
		std::cout << "ATTRIBUTES\t" << B->getAttCount() << std::endl;
		std::cout << "PARAMETERS\t" << M->getTotalParameterCount()  << std::endl;
	}
	delete OPT;
	delete L;
	delete P;
	delete B;
	delete M;
	delete G;

	cudaThreadExit();
	cout << "TIME\t\t" << ((clock()-__start)/(1.0*CLOCKS_PER_SEC)) << "s" << endl;
}

/////////////////////////////////////////////////////////////////////////////////////////////////////////
// PARSE COMMAND LINE
/////////////////////////////////////////////////////////////////////////////////////////////////////////
int parseCMD(unsigned int ac, char** av, Options& crfopts){
	std::cout << "LINEAR-CHAIN CRF@CUDA 1.0";// << std::endl;
        std::cout << ", (C) 2011 Nico Piatkowski, All rights reserved." << std::endl;
        std::cout << "SFB876 - http://sfb876.tu-dortmund.de" << std::endl << std::endl;
	//std::cout << "There is ABSOLUTELY NO WARRANTY; not even for MERCHANTABILITY" << std::endl;
        //std::cout << "or FITNESS FOR A PARTICULAR PURPOSE." << std::endl << std::endl;

	po::options_description general_options("GENERAL");
	general_options.add_options()
	    ("predict", "Don't learn. Just predict the labels of the input data. This requires a model and testing data.")
	    ("labelcount,Y", po::value<int>()->default_value(32), "Max number of labels, <=64")
	    ("batchsize,b", po::value<int>()->default_value(96), "Batchsize")
	    ("iterations,I", po::value<int>()->default_value(1), "Number of iterations")
	    ("weight,w", po::value<float>()->default_value(0.05), "Initial weight")
	    ("help,h", "this help")
	;
	po::options_description file_options("FILE HANDLING");
	file_options.add_options()
		("modelfile,M", po::value<std::string>(), "Model file")
		("labelfile,L", po::value<std::string>(), "Label output file")
		("testfile,T", po::value<std::string>(), "Testset")
	;
	po::options_description optimization_options("OPTIMIZATION");
	optimization_options.add_options()
	    ("eta,e", po::value<float>()->default_value(0.1), "Stepsize")
	;

	po::options_description misc_options("EVALUATION");
	misc_options.add_options()
	    ("prf", "Show precision, recall and F1-score after every iteration")
	    ("iob", "Use IOB-Tagging")
	    ("clear", "Clear screen after every iteration")
	;

	po::options_description cuda_options("CUDA");
	cuda_options.add_options()
	    ("device", po::value<int>()->default_value(0), "Choose CUDA device")
	;

	po::options_description cmdline_options;
	cmdline_options.add(general_options).add(file_options).add(optimization_options).add(misc_options).add(cuda_options);

	po::variables_map vm;
	po::store(po::parse_command_line(ac, av, cmdline_options), vm);
	po::notify(vm);    
	
	if (vm.count("help")) {
		std::cout << "Usage: " << av[0] << " [Options] < TRAININGDATA";
	    std::cout << cmdline_options << std::endl;
	    return 1;
	}

	int deviceCount = 0;
	if (cudaGetDeviceCount(&deviceCount) != cudaSuccess) {
		std::cout << "NO CUDA CAPABLE DEVICE DETECTED!" << std::endl;
		return 1;
	}

	int device = 0;

	if (vm.count("device")){
		device = vm["device"].as<int>();
	}

	if (deviceCount <= device) {
		std::cout << "NO CUDA CAPABLE DEVICE WITH ID " << device << " DETECTED! " << std::endl;
		return 1;
	}

	cudaDeviceProp deviceProp;
    	cudaGetDeviceProperties(&deviceProp, device);
	std::cout << "DEVICE\t\t" << deviceProp.name << " (Device " << device << ")" << std::endl;

	const int maxThreadsPerBlock = deviceProp.maxThreadsPerBlock;

	if (vm.count("labelfile")){
		crfopts.label_file = vm["labelfile"].as< std::string >();
		crfopts.writeLabels		= true;
	}else crfopts.writeLabels	= false;

	if (vm.count("modelfile")){
		crfopts.model_file = vm["modelfile"].as< std::string >();
		crfopts.hasModel = true;
	}else crfopts.hasModel = false;

	if ( vm.count("predict")){
		crfopts.test	= true;
		crfopts.predict	= true;
	}else{
		crfopts.predict	= false;
	}

	if (vm.count("testfile")){
		crfopts.test_file = vm["testfile"].as< std::string >();
		crfopts.test		= true;
	}else{
		crfopts.test		= false;
	}

	if (vm.count("batchsize")){
		crfopts.batchsize = vm["batchsize"].as<int>();
		assert(crfopts.batchsize && crfopts.batchsize < 4096);
		if(!crfopts.predict) std::cout << "BATCHSIZE\t" << crfopts.batchsize << std::endl;
	}

	if (vm.count("iterations")){
		crfopts.iterations = vm["iterations"].as<int>();
	}
	crfopts.nodes = 128;
	crfopts.o = O_SGD;

	if (vm.count("eta")){
		crfopts.stepsize = vm["eta"].as<float>();
		if(!crfopts.predict) std::cout << "ETA\t\t" << crfopts.stepsize << std::endl;
	}
	if (vm.count("weight")){
		crfopts.weight = vm["weight"].as<float>();
		if(!crfopts.predict) std::cout << "INITIAL-WEIGHT\t" << crfopts.weight << std::endl;
	}
	crfopts.blockBase      = 64;
	crfopts.threadBase     = 128;
	crfopts.blockBaseBP    = 32;
	crfopts.threadBaseBP   = 16;
	crfopts.threadBaseBP2  = 2;
	crfopts.tieParameters  = true;
	crfopts.g              = G_CHAIN;
	crfopts.e              = E_FB;

	if (vm.count("labelcount")){
		crfopts.labelcount = vm["labelcount"].as<int>();
		assert(crfopts.labelcount>0 && crfopts.labelcount<65);
	}

	if ( vm.count("iob")){
		crfopts.iobLabels = true;
	}else{
		crfopts.iobLabels = false;
	}

	if ( vm.count("clear")){
		crfopts.clearScreen	= true;
	}else{
		crfopts.clearScreen	= false;
	}

	if ( vm.count("prf")){
		crfopts.showPRF	= true;
	}else{
		crfopts.showPRF	= false;
	}
	crfopts.showProgress = true;
	crfopts.d = D_FLOAT;

	return 0;
}

/////////////////////////////////////////////////////////////////////////////////////////////////////////
// TESTING (FROM RAM)
/////////////////////////////////////////////////////////////////////////////////////////////////////////
template <class T>
void readTestset(Options& crfopts, unsigned int &tests, Model<T>* M, ObservationSet<T>* B, std::vector<ObservationSet<T>* >&testsets){
	clock_t start = clock();
	ifstream dataFile(crfopts.test_file.c_str());
	ObservationSet<T>* test = new ObservationSet<T>(B->getAttDict(), B->getClassDict(), M), *nextTest;

	while(!dataFile.eof()){
		test->readFromStream(&dataFile, crfopts.batchsize);
		assert(test->getLabelCount() <= crfopts.labelcount);
		tests += test->getSize();
		if(crfopts.showProgress) std::cout << "\rLOADING TESTSET\t" << tests << std::flush;
		M->init(test);
		M->store(test);
		testsets.push_back(test);
		test = new ObservationSet<T>(B->getAttDict(), B->getClassDict(), M, false);
	}
	delete test;
	dataFile.close();
	float time = (float)(clock()-start)/(float)CLOCKS_PER_SEC;
}

bool startChunk(char p, char c, unsigned int pt, unsigned int ct){
	return (p == 'B' && c == 'B') || (p == 'I' && c == 'B') || (p == 'O' && c == 'B') || (p == 'O' && c == 'I') || (c != 'O' && pt != ct);
}

bool endChunk(char p, char c, unsigned int pt, unsigned int ct){
	return (p == 'B' && c == 'B') || (p == 'B' && c == 'O') || (p == 'I' && c == 'B') || (p == 'I' && c == 'O') || (c != 'O' && pt != ct);
}

template <class T>
void evaluate(Options& crfopts, unsigned int itt, unsigned int total, unsigned int total_tests, float trainingTime, Model<T>* M, InferenceEngine<T>* P, ObservationSet<T>* B, std::vector<ObservationSet<T>* > &testsets){

	std::cout << std::endl;

  	clock_t a = clock();

	ofstream outFile;

	const unsigned int labelCharWidth = 13;

	const std::string writePrefix = "_ _ ";
	const std::string writeSeperator = " ";

	unsigned int rc=0, mc=0, pc=0;
	unsigned int rcounts[crfopts.labelcount], mcounts[crfopts.labelcount], pcounts[crfopts.labelcount];

	bool iobScan = false;

	unsigned int tokenCounter = 0, correctTags = 0, tests = 0;

	float testingTime = 0.0f;
	float prec = 0.0f, rec = 0.0f, f = 0.0f;

	memset(rcounts, 0, crfopts.labelcount*sizeof(int));
	memset(mcounts, 0, crfopts.labelcount*sizeof(int));
	memset(pcounts, 0, crfopts.labelcount*sizeof(int));
	rc=0;
	mc=0;
	pc=0;

	testingTime = 0;

	typename std::vector<ObservationSet<T>* >::const_iterator tI = testsets.begin();
	ObservationSet<T>* test;

	std::string *label;

	bool w = crfopts.writeLabels && (itt+1 == crfopts.iterations);

	if(w) outFile.open(crfopts.label_file.c_str());

	std::string OO("O");

	for(; tI != testsets.end(); tI++){

		tests += (*tI)->getSize();

		if(crfopts.showProgress) std::cout << "\rTESTING \t" << tests << "/" << total_tests << std::flush;
		a = clock();
		test = *tI;
		M->init(test);
		unsigned int* reference  = (unsigned int*)malloc((test->getNodeMapSize()-1)*sizeof(unsigned int));
		memcpy(reference, test->getLabels(), (test->getNodeMapSize()-1)*sizeof(unsigned int));
		unsigned int* prediction = P->computeMaxPropAssignment(test);
		testingTime += (float)(clock()-a)/(float)CLOCKS_PER_SEC;

		iobScan = false;

		char lastCorrect, correct, lastGuessed, guessed;
		unsigned int guessedType, correctType, lastGuessedType, lastCorrectType;

		for(unsigned int qq=0; qq<(*tI)->getSize(); qq++){
			for(unsigned int pp=0; pp<((*tI)->getStructureSizes())[qq]; pp++){
			unsigned int i = (((*tI)->getStructureMap())[qq])+pp;

			if(w) outFile << writePrefix << *B->getClassDict()->getToken(reference[i]) << writeSeperator << *B->getClassDict()->getToken(prediction[i]) << std::endl;

			if(!crfopts.iobLabels){
				if(reference[i] == prediction[i]){
					mc++;
					mcounts[prediction[i]]++;
				}
				rc++;
				rcounts[reference[i]]++;
				pcounts[prediction[i]]++;

			}else{

				if(!i){
					lastCorrect = 'O';
					lastGuessed = 'O';

					lastCorrectType = B->getClassDict()->getTokenId(&OO);
					lastGuessedType = B->getClassDict()->getTokenId(&OO);
				}

				correct = B->getClassDict()->getToken(reference[i])->at(0);
				guessed = B->getClassDict()->getToken(prediction[i])->at(0);

				std::string correctTypeS(*B->getClassDict()->getToken(reference[i]));
				std::string guessedTypeS(*B->getClassDict()->getToken(prediction[i]));

				if(correct != 'O') correctTypeS.at(0) = 'B';
				if(guessed != 'O') guessedTypeS.at(0) = 'B';

				correctType = B->getClassDict()->getTokenId(&correctTypeS);
				guessedType = B->getClassDict()->getTokenId(&guessedTypeS);

				if(iobScan){
					if(endChunk(lastCorrect, correct, lastCorrectType, correctType) &&
					   endChunk(lastGuessed, guessed, lastGuessedType, guessedType) &&
					   (lastGuessedType == lastCorrectType)){

						iobScan = false;

						mc++;
						mcounts[lastCorrectType]++;

					}else if((endChunk(lastCorrect, correct, lastCorrectType, correctType) !=
							  endChunk(lastGuessed, guessed, lastGuessedType, guessedType)) ||
							  (guessedType != correctType)){

						iobScan = false;
					}
				}

				if(startChunk(lastCorrect, correct, lastCorrectType, correctType) &&
                	startChunk(lastGuessed, guessed, lastGuessedType, guessedType) &&
					guessedType == correctType){
					iobScan = true;
				}

				if(startChunk(lastCorrect, correct, lastCorrectType, correctType)){
					rc++;
					rcounts[correctType]++;
				}

				if(startChunk(lastGuessed, guessed, lastGuessedType, guessedType)){
					pc++;
					pcounts[guessedType]++;
				}

				if ( correct == guessed && guessedType == correctType ) { 
		        	correctTags++; 
      			}

				tokenCounter++;

			    lastGuessed = guessed;
			    lastCorrect = correct;
			    lastGuessedType = guessedType;
			    lastCorrectType = correctType;
			}
			}
		}

		if(crfopts.iobLabels && iobScan){
			mc++;
			mcounts[lastCorrectType]++;
		}

		free(reference);
		free(prediction);
		M->store(*tI);
	}

	if(w) outFile.close();
	if(crfopts.clearScreen) int r = system("clear");
	if(crfopts.showPRF){
		std::cout << std::endl << std::endl << std::setw(labelCharWidth) << rc << " phrases\t" << pc << " found\t" << mc << " correct" << std::endl;
		std::cout << std::endl << std::setw(labelCharWidth) << "LABEL" << '\t' << "PREC" << '\t' << "REC" << '\t' << "F1" << std::endl;
		std::cout << std::setw(labelCharWidth) << "-----" << '\t' << "----" << '\t' << "---" << '\t' << "--" << std::endl;

		if(crfopts.iobLabels){
			prec = 100.0f*(float)mc/(float)pc;
			rec = 100.0f*(float)mc/(float)rc;
			f = (2.0f*prec*rec)/(prec+rec);
			std::cout << std::setw(labelCharWidth) << "ALL" << setprecision(4) << '\t' << prec << '\t' << rec << '\t' << f << '\t' << std::endl;
		}

		for(unsigned int i=0; i<crfopts.labelcount; i++){
			prec = 100.0f*(float)mcounts[i]/(float)pcounts[i];
			rec = 100.0f*(float)mcounts[i]/(float)rcounts[i];
			f = (2.0f*prec*rec)/(prec+rec);
			if(!crfopts.iobLabels) std::cout << std::setw(labelCharWidth) << *B->getClassDict()->getToken(i) << setprecision(4) << '\t' << prec << '\t' << rec << '\t' << f << std::endl;
			else{
				label = B->getClassDict()->getToken(i);
				if(label->at(0) == 'B') std::cout << std::setw(labelCharWidth) << label->substr(2, label->size()-2) << setprecision(4) << '\t' << prec << '\t' << rec << '\t' << f << '\t' << std::endl;
			}
		}
	}

	if(crfopts.showProgress) std::cout << std::endl;
	if(crfopts.iobLabels){
		 std::cout << "ACCURACY\t"<< (100.0f*(float)correctTags/(float)tokenCounter) <<  std::endl;
	}else{
		 std::cout << "ACCURACY\t"<< (100.0f*(float)mc/(float)rc) <<  std::endl;
	}
}
/////////////////////////////////////////////////////////////////////////////////////////////////////////

